UNKNOWN = "not known"
#Workflow class used to store information of the workflow being processed on
class Workflow:
	def __init__(self, name, workflow, link="local"):
		self.name = name;
		self.inputs = [] if not "inputs"  in workflow else self.createInputOutputArray(workflow["inputs"])
		self.outputs = [] if not "outputs" in workflow else self.createInputOutputArray(workflow["outputs"])
		self.steps = [] if not "steps" in workflow else self.createStepArray(workflow["steps"])
		self.link = link
		self.parent = []

	def createStepArray(self, steps) :
		import cwlbrowser.step as s
		temp = []
		#print (steps)
		if (isinstance(steps, dict)) :
			for key, value in steps.items() :
				stepObj = s.Step(value, name=key)
				temp.append(stepObj)
		elif(isinstance(steps, list)) :
			for step in steps :
				stepObj = s.Step(step)
				temp.append(stepObj)
		else :
			temp = []
		return temp

	def loadStep(self, run) :
		import cwlbrowser.browser as c
		import cwlbrowser.util as util
		if run not in self.getStepsByRun():
			raise Exception('{} is not a step in {}'.format(run, self.name))
		else :
			stepLink = util.constructLink(self.link, run)
			step = c.loadWithLink(stepLink)
			step.setParent(self)
			return step



	def createInputOutputArray(self, elements) :
		import cwlbrowser.util as util
		return util.createInputOutputArray(elements, False)

	def setParent(self, parent) :
		self.parent = parent


	def getInputsByName(self) :
		return self.getElementsByAttribute(self.inputs, "name")

	def getInputsByType(self) :
		return self.getElementsByAttribute(self.inputs, "type")

	def getOutputsByName(self) :
		return self.getElementsByAttribute(self.outputs, "name")

	def getOutputsByType(self) :
		return self.getElementsByAttribute(self.outputs, "type")

	def getStepsByName(self) :
		return self.getElementsByAttribute(self.steps, "name")

	def getStepsByRun(self) :
		return self.getElementsByAttribute(self.steps, "run")

	def getElementsByAttribute(self, list_, attribute) :
		temp = []
		if list_ != [] :
			for item in list_ :
				if attribute == "name":
					temp.append(item.name)
				elif attribute == "type" :
					temp.append(item.type)
				else  :
					temp.append(item.run)
		else :
			temp = []
		return temp






#Step object used to store information of steps in the workflow
import cwlbrowser.util as util
import cwlbrowser.input_output as io
UNKNOWN = "not known"


class Step:
	def __init__(self, stepData, name="step"):
		self.name = name
		self.inputs = []
		self.outputs = []
		self.run = []
		if(isinstance(stepData, dict)) :
			if "id" in stepData :
				self.name = stepData["id"]
			if "in" in stepData :
				self.inputs = self.createInputOutputArray(stepData["in"])
			if "out" in stepData :
				self.outputs  = self.createInputOutputArray(stepData["out"])
			if "run" in stepData :
				self.run = stepData["run"]

	def createInputOutputArray(self, elements) :
		return util.createInputOutputArray(elements, True)

		
	def getInputsByName(self) :
		temp = []
		for input_ in self.inputs :
			temp.append(input_.name)
		return temp

	def getInputsByType(self) :
		temp = []
		for input_ in self.inputs :
			temp.append(input_.type)
		return temp

	def getOutputsByName(self) :
		temp = []
		for output in self.outputs :
			temp.append(output.name)
		return temp


	def getOutputsByType(self) :
		temp = []
		for output in self.outputs :
			temp.append(output.type)
		return temp

	def __str__(self):
		return "Step: {0} | Run: {1}".format(self.name, self.run)


#Basic similarity checker
#Carries out bottom level comparison of two workflows depending on a certain attribute
# IE inputs outputs steps
IO_WEIGHTING = 15
STEP_WEIGHTING = 70
class BasicSimilarityChecker:
	def __init__(self) :
		self.overallMatch = 0
		self.attribute = []
		self.workflow1= []
		self.workflow2 = []
		self.fullListWorkflow1 = []
		self.fullListWorkflow2 = []
		self.differencesWorkflow1 = {}
		self.differencesWorkflow2 = {}
		self.matchingItems = []

	#sets up two lists to carry out comparisons between the corresponding workflows
	def compare(self, workflow1, workflow2, attribute) :
		self.attribute = attribute
		self.workflow1 = workflow1
		self.workflow2 = workflow2
		if (attribute == "inputs") :
			self.fullListWorkflow1 = workflow1.getInputsByName()
			self.fullListWorkflow2 = workflow2.getInputsByName()
		elif (attribute == "outputs") :
			self.fullListWorkflow1 = workflow1.getOutputsByName()
			self.fullListWorkflow2 = workflow2.getOutputsByName()
		else :
			self.fullListWorkflow1 = workflow1.getStepsByName()
			self.fullListWorkflow2 = workflow2.getStepsByName()
		self.similarityCheck(self.fullListWorkflow1, self.fullListWorkflow2)
		
	#Evaluates metrics and carries out the actual comparison
	def similarityCheck(self, workflow1Attributes, workflow2Attributes):
		if self.attribute == "steps" :
			weighting = STEP_WEIGHTING
		else :
			weighting = IO_WEIGHTING
		self.differencesWorkflow1 = set(workflow1Attributes).difference(workflow2Attributes)
		self.differencesWorkflow2 = set(workflow2Attributes).difference(workflow1Attributes)
		if len(workflow1Attributes) >= len(workflow2Attributes) :
			matchingItems = set(workflow2Attributes).intersection(workflow1Attributes)
			denominator = len(workflow1Attributes)
		else :
			matchingItems = set(workflow1Attributes).intersection(workflow2Attributes)
			denominator = len(workflow2Attributes)
		if denominator <= 0 :
			denominator = 1
		self.matchingItems = matchingItems
		self.overallMatch = len(matchingItems) / denominator * weighting

		
	def getOverallMatch(self) :
		return self.overallMatch

	def getDifferingItems(self) :
		differingItems = []
		for i in self.differencesWorkflow1 :
			differingItems.append(i)
		for i in self.differencesWorkflow2 :
			differingItems.append(i)
		return differingItems


#Similarity checker object
#Consists of three similarity objects that compare a certain
#attribute of two workflows (inputs, outputs, steps)

STEPS_WEIGHTING = 70
IO_WEIGHTING = 15

class SimilarityChecker :
	def __init__(self) :
		import cwlbrowser.basic_similaritychecker as b
		self.overallMatch = 0
		self.workflow1 = []
		self.workflow2 = []
		self.inputSimilarityChecker = b.BasicSimilarityChecker()
		self.stepSimilarityChecker = b.BasicSimilarityChecker()
		self.outputSimilarityChecker = b.BasicSimilarityChecker()

	def similarityCheck(self, workflow1, workflow2) :
		self.workflow1 = workflow1
		self.workflow2 = workflow2
		self.inputSimilarityChecker.compare(workflow1, workflow2, "inputs")
		self.stepSimilarityChecker.compare(workflow1, workflow2, "steps")
		self.outputSimilarityChecker.compare(workflow1, workflow2, "outputs")
		self.overallMatch = round((self.inputSimilarityChecker.overallMatch 
							+ self.outputSimilarityChecker.overallMatch 
							+ self.stepSimilarityChecker.overallMatch), 3)

	def getOverallMatch(self) :
		return self.overallMatch



#Main central python module used to carry out browsing of workflows on Jupyter notebook
import requests
import base64
import yaml
import os
import cwltool
import subprocess
import cwlbrowser.workflow as wf
import cwlbrowser.util as util
import cwlbrowser.similaritychecker as s
from IPython.display import SVG, display, HTML
import time
import itertools
import os


out_workflow = {}
steps = []
STEPS_WEIGHTING = 70
IO_WEIGHTING = 15
workflowlink = []


#retrieves .cwl file from provided github link
def retrieveFileThruLink(link, path):
	req = requests.get(link)
	if req.status_code != 200:
	    print(req.status_code)
	    if req.status_code == 404:
	    	print ("cwl file not found. Maybe due to incorrect url")
	    	print("\n")
	else:
		req = req.json()  	    
		content = base64.b64decode(req['content'])
		content = content.decode("utf-8")
		try:
			out_workflow = yaml.safe_load(content)
			#print (out_workflow)
			return wf.Workflow(path, out_workflow, workflowlink)
		except (yaml.YAMLError) as yamlError:
			print ("Error in loading the cwl file")
			print (yamlError)

#main load method.
#path to local file or github link to file can be passed as an argument
def load(workflow, link=True):
	if(link == False):
		global out_workflow
		if os.path.exists(workflow):
			file = open(workflow, 'rb')
			try:
				out_workflow = yaml.safe_load(file)
				return wf.Workflow(workflow, out_workflow)
			except (yaml.YAMLError):
				print ("Error in loading the cwl file")
		else:
			print("File not located")
			return -1;
	else:
		return loadWithLink(workflow)

#pre processes link so that it is in correct format to pass to 
#retrieveFileThruLink method
def loadWithLink(link) :
	global workflowlink
	workflowlink = link
	lhs, rhs = link.split("/blob/", 1)
	branch, path = rhs.split("/", 1)
	ownerrepo = lhs.replace('https://github.com', "")
	finallink = "https://api.github.com/repos" + ownerrepo + "/contents/" + path
	return retrieveFileThruLink(finallink, path)

#print the specified attribute you want 
#pass the attribute as a string to the method
#E.G. printWorfklow("inputs")
def printWorkflowAttr(attribute) :
	if attribute in out_workflow.keys() :
		print(attribute + ":")
		for item in out_workflow[attribute] :
			print (item)
		print ("\n")
	else :
		print("No such attribute as: " + attribute + " found in workflow")


#displays visual representation of workflow using CWL Viewer API
def displayGraph(workflow, type="link"):
	if (type  == "link") :
		BASE_URL = 'https://view.commonwl.org'

		HEADERS = {
		    'user-agent': 'my-app/0.0.1',
			'accept': 'application/json'
		}

		addResponse = requests.post(BASE_URL + '/workflows', 
								data={'url': workflow.link},		
								headers=HEADERS)
		if (addResponse.status_code == 200) :
			postExistingWorkflowGraph(workflow.link)
		elif (addResponse.status_code == requests.codes.accepted):
			qLocation = addResponse.headers['location']

			# Get the queue item until success
			qResponse = requests.get(BASE_URL + qLocation,
									headers=HEADERS,
									allow_redirects=False)
			maxAttempts = 5
			while qResponse.status_code == requests.codes.ok and qResponse.json()['cwltoolStatus'] == 'RUNNING' and maxAttempts > 0:
				time.sleep(5)
				qResponse = requests.get(BASE_URL + qLocation,
										headers=HEADERS,
										allow_redirects=False)
				maxAttempts -= 1
			if 'location' in qResponse.headers:
			# Success, get the workflow
				workflowResponse = requests.get(BASE_URL + qResponse.headers['location'], headers=HEADERS)
				if (workflowResponse.status_code == requests.codes.ok):
					workflowJson = workflowResponse.json()
					# Do what you want with the workflow JSON
					# Include details in documentation files etc
					display(SVG(BASE_URL + workflowJson['visualisationSvg']))
				else:
					print('Could not get returned workflow')
			elif (qResponse.json()['cwltoolStatus'] == 'ERROR'):
				# Cwltool failed to run here
				print(qResponse.json()['message'])
			elif (maxAttempts == 0):
				print('Timeout: Cwltool did not finish')
			else :
				print('Something is not right')

		else:
			print (addResponse.status_code)
			print('Error adding workflow')
	elif (type== "file"):
		name, suffix = workflow.split(".cwl") 
		p = subprocess.run(["cwltool", "--print-dot", workflow], capture_output=True)
		if(p.returncode != 0) :
			print ("Something went wrong")
			print(bytes.decode(p.stderr))
		else :
			graphFile = name + ".svg"
			display(SVG(graphFile))
	else :
		print("Invalid type")

	"""
	"""

#displays graph of workflow if it has already been uploaded to website
def postExistingWorkflowGraph(link) :
	BASE_URL = 'https://view.commonwl.org'

	HEADERS = {
		'accept': 'application/json'
	}
	shortenedLink = link.replace("https://", "")
	finishedlink = BASE_URL + "/workflows/" + shortenedLink
	req = requests.get(finishedlink, headers=HEADERS)
	if req.status_code != 200:
		print(req.status_code)
	else:
		req = req.json()     
		display(SVG(BASE_URL + req['visualisationSvg']))

#displays tabular representation of workflow
#specific attributes can be displayed on their own
def displayTables(workflow, attr='All') :
	caption = ""
	if attr == 'All':
		caption = "\n {} INPUTS:".format(workflow.name)
		tabulateWorkflow(workflow.inputs, caption)
		caption = "\n {} OUTPUTS:".format(workflow.name)
		tabulateWorkflow(workflow.outputs, caption)
		caption = "\n {} STEPS:".format(workflow.name)
		tabulateWorkflow(workflow.steps, caption, step=True)

	elif attr == 'inputs' or attr == 'input' :
		caption = "\n {} INPUTS:".format(workflow.name)
		tabulateWorkflow(workflow.inputs, caption)

	elif attr == 'outputs' or attr == 'output':
		caption = "\n {} OUTPUTS:".format(workflow.name)
		tabulateWorkflow(workflow.outputs, caption)

	elif attr == 'steps' or attr =='step' :
		caption = "\n {} STEPS:".format(workflow.name)
		tabulateWorkflow(workflow.steps, caption, step=True)

	else:
		print("Invalid attribute specified: Please use 'outputs', 'inputs', 'steps' or 'All")


#Displays results of similarity checker
def displayStats(similarityChecker) :
	tabulateSimChecker(similarityChecker.inputSimilarityChecker)
	tabulateSimChecker(similarityChecker.outputSimilarityChecker)
	tabulateSimChecker(similarityChecker.stepSimilarityChecker)
	print("Overall match: {}".format(similarityChecker.overallMatch))

def tabulateSimChecker(similarityChecker) :
	caption = '<caption style="font-weight:bold; font-size: 20px">{} and {} {}</caption>'.format(similarityChecker.workflow1.name, 
														similarityChecker.workflow2.name, similarityChecker.attribute)
	data = ""
	tableBody = setColours(similarityChecker.fullListWorkflow1, similarityChecker.fullListWorkflow2,
							similarityChecker.differencesWorkflow1, 
							similarityChecker.differencesWorkflow2, similarityChecker.attribute)
	data = data + tableBody
	display(HTML('<table style="width:100%">{}<tr><th>{}</th><th>{}</th></tr>{}</table>'.format(caption, similarityChecker.workflow1.name, similarityChecker.workflow2.name, data)))


#sets colours of cells in table to distinguish differing and similar elements of the two workflows being compared
def setColours(worklow1Attributes, workflow2Attributes, workflow1Diff, workflow2Diff, attr) :
	tableBody = ""
	styleI = ""
	styleJ = ""
	uncolored = 'style="border: 1px solid black"' 
	green = 'style="border: 1px solid black; background-color:limegreen"'
	diff = 'style="border: 1px solid black; background-color:{}"'.format("yellow" if not attr == 'steps' else "red")
	for i, j in itertools.zip_longest(worklow1Attributes, workflow2Attributes) :
		styleI = green if not (i in workflow1Diff) else diff
		styleJ = green if not (j in workflow2Diff) else diff
		(i, styleI) = (i, styleI)  if not(i == None) else ("", uncolored)
		(j, styleJ) = (j, styleJ) if not(j == None) else ("", uncolored)
		tableBody = tableBody + ('<tr><td {}>{}</td><td {}>{}</td></tr>'.format(styleI, i, styleJ, j))
	return tableBody


#Helper method used to display tabular view of workflow via HTML
def tabulateWorkflow(list_, inputCaption, step=False) :
	caption = '<caption style="font-weight:bold; font-size: 20px">{}</caption>'.format(inputCaption)
	page = "<html>" 
	style = 'style= "border: 1px solid black"'
	data = ""
	if step == False :
		attr = "Type"
		for i in list_ :
			data = data + ("<tr {}><td {}>{}</td><td {}>{}</td></tr>".format(style, style, i.name, style, i.type))
	else :
		attr = "Run"
		for i in list_ :
			data = data + ("<tr {}><td {}>{}</td><td {}>{}</td></tr>".format(style, style, i.name, style,i.run))
	page = page + '<body><table style="width:100%;border: 1px solid black">{}<tr><th {}>Name</th><th {}>{}</th></tr>{}</table></body>'.format(caption, style, style, attr, data)
	page = page + "</html>"
	display(HTML(page))











	



#Utility module that has general purpose methods for displaying information 
def printAttr(attribute, attributeName, subjectName):
	print(subjectName + " " + attributeName + ":")
	print("-----------------")
	for item in attribute:
		if(isinstance(item, dict)):
			print(item["id"])
		else:
			print(item)
	print("\n")


def compare(name1, name2,attributeX, attributeY, attributeName):
	count1 = 0
	count2 = 0
	difference = 0
	for input_ in attributeX:
		count1 = count1 + 1
	for input_2 in attributeY:
		count2 = count2 + 1
	if(count1 >= count2):
		difference = count1 - count2
	else:
		difference = count2 - count1
	print("-------------------------------------")
	print("--------- " + attributeName + " comparison------------")
	print("-------------------------------------")
	print(name1 + "| no. of " + attributeName+ ": " + str(count1))
	print(name2 + "| no. of " + attributeName+ ": " + str(count2))
	print("Difference: " + str(difference))
	print("-------------------------------------")
	print("\n")


def createInputOutputArray(elements, step) :
	import cwlbrowser.workflow as wf
	import cwlbrowser.input_output as io
	temp =[]
	if(isinstance(elements, dict)) :
		for key, value in elements.items() :
			obj = io.InputOutput(value, step, name=key) 
			temp.append(obj)
	elif (isinstance(elements, list)) :
		for e in elements :
			obj = io.InputOutput(e, step) 
			temp.append(obj)
	else :
		temp = []
	return temp


def constructLink(parentLink, link) :
	parentlhs, parentrhs = parentLink.rsplit('/', 1)
	if('/' in link) :
		lhs, rhs = link.split('/', 1)
		if(lhs == '..') :
			root, path = parentlhs.rsplit('/', 1)
			finalLink = root + '/' + rhs
		else :
			finalLink = ""
	else :
		finalLink = parentlhs + '/' + link
	return finalLink







	




from setuptools import setup
# read the contents of README file
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='cwlbrowser',
      version='4.0',
      description='Python library that browses and analyses workflows in CWL',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://gitlab.cs.man.ac.uk/mbaxasp7/cwlbrowser',
      author='Sean Pertet',
      author_email='sean.pertet@student.manchester.ac.uk',
      packages=['cwlbrowser'],
      install_requires=[
          'pyyaml',
          'requests',
          'cwlref-runner',
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data=True,
      zip_safe=False)


#Object used to represent inputs and outputs of a workflow 
UNKNOWN = "not known"

class InputOutput:
	def __init__(self, inputData, step, name=""):
		self.name = name
		self.type = UNKNOWN
		if(isinstance(inputData, dict)) :
			if "id" in inputData :
				self.name = inputData["id"]
			if "type" in inputData :
				self.type = inputData["type"]
		elif(isinstance(inputData, str)) :
			#if name hasn't been assigned 
			if (self.name == "") :
				self.name = inputData
			elif (name != inputData) :
				if (step == False) :
					self.type = inputData
				else:
					self.type = UNKNOWN
				

	def __str__(self):
		return "Input: {0} | Type: {1}".format(self.name, self.type)